import uuid
import json
from pika import BasicProperties

from .utils import current_time
from .subscription import Subscription

class Message:
    def __init__(self):
        self.__topic = None
        self.__sid = None
        self.__body = ''
        self.__reply_to = None
        self.__correlation_id = None
        self.__content_type = None
        self.__created_at = current_time()
        self.__metadata = {}
        self.__on_reply = None
        self.__on_timeout = None
        self.__timeout_ms = None
        self.__ctag = None

    def get_properties(self):
        return BasicProperties(
            content_type=self.content_type(),
            timestamp=self.created_at(),
            headers=self.metadata(),
            reply_to=self.reply_to(),
            correlation_id=self.correlation_id(),
            expiration=self.timeout_ms()
        )

    def set_properties(self, props):
        self.set_created_at(props.timestamp)
        self.set_correlation_id(props.correlation_id)
        self.set_reply_to(props.reply_to)
        self.set_timeout_ms(props.expiration)
        self.set_content_type(props.content_type)
        headers = props.headers
        if 'rpc-status' in headers:
            rpc_status = headers['rpc-status']
            if isinstance(rpc_status, str):
                headers['rpc-status'] = json.loads(rpc_status)
        self.add_metadata(headers)

    def create_reply(self):
        rep_msg = Message()
        rep_msg.set_topic(self.reply_to())                 \
                .set_correlation_id(self.correlation_id()) \
                .set_content_type(self.content_type())
        return rep_msg

    def topic(self):
        return self.__topic
    
    def set_topic(self, topic):
        self.__topic = topic
        return self

    def reply_to(self):
        return self.__reply_to

    def set_reply_to(self, subscription):
        if isinstance(subscription, Subscription):
            self.__reply_to = subscription.name()
            self.__ctag = subscription.id()
            if self.__correlation_id == None:
                self.set_correlation_id(self.__new_cid())
        elif isinstance(subscription, str):
            self.__reply_to = subscription
            if self.__correlation_id == None:
                self.set_correlation_id(self.__new_cid())
        return self

    def has_reply_to(self):
        return self.__reply_to != None

    def consumer_tag(self):
        return self.__ctag
    
    def set_consumer_tag(self, ctag):
        self.__ctag = ctag
        return self

    def has_consumer_tag(self):
        return self.__ctag != None

    def correlation_id(self):
        return self.__correlation_id

    def set_correlation_id(self, cid=None):
        self.__correlation_id = cid if cid else self.__new_cid()
        return self

    def has_correlation_id(self):
        return self.__correlation_id != None
    
    def body(self):
        return self.__body

    def set_body(self, body):
        self.__body = body
        return self

    def content_type(self):
        return self.__content_type
    
    def set_content_type(self, content_type):
        self.__content_type = content_type
        return self

    def set_created_at(self, created_at):
        self.__created_at = created_at
        return self

    def created_at(self):
        return self.__created_at

    def metadata(self):
        return self.__metadata

    def add_metadata(self, mdata):
        self.__metadata.update(mdata)
        return self

    def set_on_reply(self, handle):
        self.__on_reply = handle
        return self

    def on_reply(self):
        return self.__on_reply

    def has_on_reply(self):
        return self.__on_reply != None

    def set_on_timeout(self, handle):
        self.__on_timeout = handle
        return self

    def on_timeout(self):
        return self.__on_timeout

    def has_on_timeout(self):
        return self.__on_timeout != None

    def set_timeout_ms(self, ms):
        if ms != None:
            self.__timeout_ms = str(ms)
        return self
    
    def timeout_ms(self):
        return self.__timeout_ms
    
    def has_timeout_ms(self):
        return self.__timeout_ms != None

    def pack(self, obj):
        self.__body = obj.SerializeToString() if obj != None else ''
        self.__content_type = 'application/x-protobuf'
        return self

    def unpack(self, pb):
        msg = pb()
        msg.ParseFromString(self.__body)
        return msg

    def __new_cid(self):
        return '{:X}'.format(uuid.uuid1().int >> 64)


from is_wire.core.channel import *
from is_wire.core.subscription import *
from is_wire.core.message import *
from is_wire.core.tracer import *
from is_wire.core.logger import *
from is_wire.core.utils import *


import os
import uuid
from datetime import datetime

def current_time():
    return int((datetime.utcnow() - datetime(1970, 1, 1)).total_seconds() * 1000)

def consumer_id():
    return '{}/{:X}'.format(os.uname()[1], uuid.uuid4().int >> 64)


import pika
import uuid
import json
import re
from collections import defaultdict
from threading import Timer

from .subscription import Subscription
from .message import Message
from .utils import current_time, consumer_id
from .logger import Logger

class Channel(object):
    def __init__(self, uri='amqp://guest:guest@localhost:5672'):
        params = pika.URLParameters(uri)
        self.__connection = pika.BlockingConnection(params)
        self.__amqp_channel = self.__connection.channel()
        self.__exchange = 'is'
        self.__queues = []
        self.__ctags = []
        self.__subscriptions = defaultdict(dict)
        self.__rpcs = defaultdict(dict)
        self.__timers = defaultdict(dict)
        self.__log = Logger(name='__channel__')
        self.__tracer = None

    def __on_message(self, channel, method, props, body):
        rkey = method.routing_key
        ctag = method.consumer_tag
        cid = props.correlation_id
        msg = Message()
        msg.set_body(body)
        msg.set_properties(props)

        was_subscription = False
        for regex, callback in self.__subscriptions[ctag].items():
            if regex.match(rkey):
                msg.set_topic(rkey)
                callback(msg, {})
                was_subscription = True
                break
        if not was_subscription:
            if rkey in self.__queues:
                if cid in self.__timers[ctag]:
                    self.__timers[ctag][cid].cancel()
                    del self.__timers[ctag][cid]
                if cid in self.__rpcs[ctag]:
                    self.__rpcs[ctag][cid](msg, {})
                    del self.__rpcs[ctag][cid]
                else:
                    self.__log.warn('Message to {} with CID {} with no callback assigned or removed due timeout', rkey, cid)

    def __queue_bind(self, queue, routing_key):
        self.__amqp_channel.queue_bind(exchange=self.__exchange, queue=queue, routing_key=routing_key)

    def __queue_unbind(self, queue, routing_key):
        self.__amqp_channel.queue_unbind(exchange=self.__exchange, queue=queue, routing_key=routing_key)
    
    def __queue_declare(self, queue):
        if not queue in self.__queues:
            self.__queues.append(queue)
            self.__amqp_channel.queue_declare(queue=queue, exclusive=False, durable=False, auto_delete=True, passive=False)

    def __basic_consume(self, queue, consumer_tag, callback=None):
        if not consumer_tag in self.__ctags:
            callback = self.__on_message if not callback else callback
            self.__amqp_channel.basic_consume(callback, queue=queue, consumer_tag=consumer_tag, no_ack=True)
            self.__ctags.append(consumer_tag)

    def __default_on_timetout(self, msg, context):
        self.__log.warn('Request to {} failed: deadline excedeed for message with cid: {}', msg.topic(), msg.correlation_id())

    def set_tracer(self, tracer):
        self.__tracer = tracer

    def tracer(self):
        return self.__tracer

    def publish(self, msg):
        if msg.has_reply_to() and msg.has_consumer_tag() and msg.has_correlation_id() and msg.has_on_reply():
            cid = msg.correlation_id()
            ctag = msg.consumer_tag()
            self.__rpcs[ctag][cid] = msg.on_reply()
            if msg.has_timeout_ms():
                on_timeout = msg.on_timeout() if msg.has_on_timeout() else self.__default_on_timetout
                def handle(msg, context):
                    cid = msg.correlation_id()
                    ctag = msg.consumer_tag()
                    if cid in self.__rpcs[ctag]:
                        del self.__rpcs[ctag][cid]
                        on_timeout(msg, context)
                timer = Timer(int(msg.timeout_ms()) / 1000.0, handle, [msg, {}])
                timer.start()
                self.__timers[ctag][cid] = timer
        
        self.__amqp_channel.basic_publish(self.__exchange, msg.topic(), msg.body(), msg.get_properties())

    def listen(self):
        self.__amqp_channel.start_consuming()


from opencensus.trace import tracer as tracer_module
from opencensus.trace.exporters.zipkin_exporter import ZipkinExporter
from opencensus.trace.span_context import SpanContext
from opencensus.trace.span import format_span_json

import opencensus.trace.span_context
import re
opencensus.trace.span_context.TRACE_ID_PATTERN = re.compile('[0-9a-f]{16}?')
opencensus.trace.span_context._INVALID_TRACE_ID = '0' * 16

class ZipkinTracer:
    def __init__(self, host_name='localhost', port=9411, service_name='my_service'):
        self.exporter = ZipkinExporter(
            host_name=host_name, port=port, service_name=service_name)
        self.__tracers = {}

    def start_span(self, span_name, context={}):
        if 'x-b3-traceid' in context and 'x-b3-spanid' in context:
            trace_id = str(context['x-b3-traceid'])
            span_id = str(context['x-b3-spanid'])
            span_context = SpanContext(trace_id=trace_id, span_id=span_id)
        else:
            span_context = None
        tracer = tracer_module.Tracer(exporter=self.exporter, span_context=span_context)
        span = tracer.start_span(span_name)
        span_id = format_span_json(span)['spanId']
        trace_id = tracer.span_context.trace_id
        self.__tracers[span_id] = tracer
        return self.__trace_wire(trace_id, span_id)

    def end_span(self, span):
        span_id = span['x-b3-spanid']
        if span_id in self.__tracers:
            tracer = self.__tracers[span_id]
            tracer.end_span()
            del self.__tracers[span_id]

    def __trace_wire(self, trace_id, span_id, parents_id='0000000000000000'):
        return {
            'x-b3-sampled': '1',
            'x-b3-traceid': trace_id,
            'x-b3-flags': '0',
            'x-b3-spanid': span_id,
            'x-b3-parentspanid': parents_id
        }

    def interceptor(self, span_name):
        def interceptor_dec(f):
            def wrapper(msg, context):
                span = self.start_span(span_name, context=msg.metadata())
                context.update(span)
                f(msg, context)
                self.end_span(span)
            return wrapper
        return interceptor_dec


import sys
import logging
from uuid import uuid4
from datetime import datetime
from termcolor import colored

from .utils import current_time

''' based on: https://gist.github.com/brainsik/1238935 '''
class ColorLog(object):

    colormap = dict(
        info=dict(color='white'),
        warn=dict(color='yellow'),
        error=dict(color='red'),
        critical=dict(color='white', on_color='on_red', attrs=['bold']),
    )

    def __init__(self, logger):
        self._log = logger

    def __getattr__(self, name):
        if name in ['info', 'warn', 'error', 'critical']:
            return lambda s, *args: getattr(self._log, name)(
                colored(s, **self.colormap[name]), *args)

        return getattr(self._log, name)

''' based on: https://stackoverflow.com/questions/6290739/python-logging-use-milliseconds-in-time-format'''
class MyFormatter(logging.Formatter):
    converter = datetime.fromtimestamp
    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = ct.strftime(datefmt)
        else:
            t = ct.strftime('%d-%m-%Y %H:%M:%S')
            s = '%s:%03d' % (t, record.msecs)
        return s

class Logger:
    def __init__(self, name='__name__', level=logging.INFO):
        self.LOGGING_FMT = '[%(levelname)s][%(process)d][%(asctime)s] %(message)s'
        self.log = ColorLog(logging.getLogger(name='{}{:X}'.format(name, uuid4().int >> 64)))
        formatter = MyFormatter(fmt=self.LOGGING_FMT)
        console = logging.StreamHandler()
        self.log.addHandler(console)
        console.setFormatter(formatter)
        self.log.setLevel(level)

    def info(self, formatter, *args):
        self.log.info(formatter.format(*args))

    def warn(self, formatter, *args):
        self.log.warn(formatter.format(*args))

    def error(self, formatter, *args):
        self.log.error(formatter.format(*args))
    
    def critical(self, formatter, *args):
        self.log.critical(formatter.format(*args))
        sys.exit(-1)


from .utils import consumer_id
import re

class Subscription:
    def __init__(self, channel, queue=None):
        self.__channel = channel
        self.__ctag = consumer_id()
        self.__queue = queue if queue else self.__ctag
        self.__channel._Channel__queue_declare(queue=self.__queue)
        self.__channel._Channel__queue_bind(queue=self.__queue, routing_key=self.__queue)
        self.__channel._Channel__basic_consume(queue=self.__queue, consumer_tag=self.__ctag)

    def __make_re(self, topic):
        return re.compile(re.sub('[*]', '.+', re.sub('[.]', '\.', topic)))

    def name(self):
        return self.__queue

    def id(self):
        return self.__ctag

    def subscribe(self, topic, fn):
        self.__channel._Channel__queue_bind(
            queue=self.__queue, routing_key=topic)
        self.__channel._Channel__subscriptions[self.__ctag][self.__make_re(topic)] = fn

    def unsubscribe(self, topic):
        self.__channel._Channel__queue_unbind(queue=self.__queue, routing_key=topic)
        re_topic = self.__make_re(topic)
        if re_topic in self.__channel._Channel__subscriptions[self.__ctag]:
            del self.__channel._Channel__subscriptions[self.__ctag][re_topic]


from ..core import Logger
from ..core.utils import current_time

class LogInterceptor:
    def __init__(self):
        self.log = Logger(name='__log_interceptor__')

    def before_call(self, context):
        context['t_start'] = current_time()
        return context

    def after_call(self, context):
        t_start = context['t_start']
        t_end = current_time()
        took = t_end - t_start
        service_name = context['service_name']
        if 'rpc-status' in context:
            status = context['rpc-status']['code']
            why = context['rpc-status']['why'] if 'why' in context['rpc-status'] else ''
            if status.lower() == 'ok':
                self.log.info('{};{}ms;{}', service_name, took, status)
            elif status.lower() == 'internal_error':
                self.log.error('{};{}ms;{};\'{}\'',
                               service_name, took, status, why)
            else:
                self.log.warn('{};{}ms;{};\'{}\'',
                              service_name, took, status, why)
        return context

    def interceptor(self, service_name):
        def interceptor_dec(f):
            def wrapper(msg, context):
                t_start = current_time()
                f(msg, context)
                t_end = current_time()
                took = t_end - t_start
                metadata = msg.metadata()
                if 'rpc-status' in metadata:
                    status = metadata['rpc-status']['code']
                    why = metadata['rpc-status']['why'] if 'why' in metadata['rpc-status'] else ''
                    if status.lower() == 'ok':
                        self.log.info('{};{}ms;{}', service_name, took, status)
                    elif status.lower() == 'internal_error':
                        self.log.error('{};{}ms;{};\'{}\'',
                                       service_name, took, status, why)
                    else:
                        self.log.warn('{};{}ms;{};\'{}\'',
                                      service_name, took, status, why)
            return wrapper
        return interceptor_dec


import sys

from ..core import Channel, Subscription, Message
from ..core.utils import current_time

class ServiceProvider:
    def __init__(self, channel):
        self.__channel = channel
        self.__interceptors = []
       
    def __deadline_exceeded(self, msg):
        if msg.has_timeout_ms():
            dt = current_time() - msg.created_at()
            return not dt < int(msg.timeout_ms())
        else:
            return False

    def add_interceptor(self, interceptor):
        self.__interceptors.append(interceptor)

    def delegate(self, service, request_pb, reply_pb, service_impl):
        subscription = Subscription(self.__channel, service)
        def wrapper(request_msg, context):
            context['service_name'] = service
            for intc in self.__interceptors:
                context = intc.before_call(context)
            
            request = request_msg.unpack(request_pb)
            reply = None
            status = {'code': '', 'why': ''}
            if request:
                try:
                    if not self.__deadline_exceeded(request_msg):
                        reply, status = service_impl(request)
                except Exception as ex:
                    status['code'] = 'INTERNAL_ERROR'
                    status['why'] = 'Service \'{}\' throwed an exception: \'{}\''.format(service, ex)
                except:
                    status['code'] = 'INTERNAL_ERROR'
                    status['why'] = 'Service \'{}\' throwed unkown exception of type \'{}\''.format(service, sys.exc_info()[0])
            else:
                status['code'] = 'FAILED_PRECONDITION' 
                status['why'] = 'Expected request type \'{}\' but received something else'.format(request_pb.DESCRIPTOR.full_name)
            
            timeouted = self.__deadline_exceeded(request_msg)
            if timeouted:
                status['code'] = 'DEADLINE_EXCEEDED'
                status['why'] = ''
                     
            context['rpc-status'] = status
            for intc in self.__interceptors:
                context = intc.after_call(context)
            
            if not timeouted:
                reply_msg = request_msg.create_reply()
                reply_msg.pack(reply).add_metadata({'rpc-status': status})
                self.__channel.publish(reply_msg)

        on_request = wrapper
        tracer = self.__channel.tracer()
        if tracer:
            on_request = tracer.interceptor(service)(on_request)
    
        subscription.subscribe(service, on_request)


from setuptools import setup, find_packages

setup(
    name='is_wire',
    version='1.0.5.1',
    description='',
    url='http://github.com/labviros/is-wire-py',
    author='labviros',
    license='MIT',
    packages=[ 'is_wire', 'is_wire.core', 'is_wire.rpc' ],
    zip_safe=False,
    install_requires=['pika==0.12.0', 'termcolor', 'is-msgs', 'is-opencensus']
    # change 'is-opencensus' to 'opencensus' when pull request accepted
)
